-- https://github.com/moody/DethsBagLib

local LibName, LibVersion = "DethsBagLib", "1.1"
local PREFIX = format("%s_%s_", LibName, LibVersion)

-- DethsLibLoader
local DBL = DethsLibLoader:Create(LibName, LibVersion)
if not DBL then return end

-- Upvalues
local assert, next, pairs, type, select, tremove =
      assert, next, pairs, type, select, table.remove

local RETRIEVING_ITEM_INFO = RETRIEVING_ITEM_INFO
local BACKPACK_CONTAINER, NUM_BAG_SLOTS = BACKPACK_CONTAINER, NUM_BAG_SLOTS
local GetItemInfo, GetContainerItemInfo, GetDetailedItemLevelInfo =
      GetItemInfo, GetContainerItemInfo, GetDetailedItemLevelInfo
local GetContainerNumSlots, GetContainerItemID =
      GetContainerNumSlots, GetContainerItemID

-- _DBL - private library table
local _DBL = {
  IsUpToDate = false,
  UpdateQueued = false,
  UpdateTimer = 0,
  
  Bags = {
    --[[ References to BagIndex tables should never change.
    [BagIndex] = {
      [SlotIndex] = item returned from _DBL:GetItem()
      ...
    }
    ...
    --]]
  },
  Listeners = {
    -- [function] = true
  },

  ItemLevelQueue = {
    -- [item] = true
  }
}

-- Add tables to _DBL.Bags
for bag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do _DBL.Bags[bag] = {} end

--[[ Debug func
local function debug(msg) print(format("|cFFFF7F50[DBL]|r %s", msg)) end
--]] local debug = nop

-- ============================================================================
-- Helper Functions
-- ============================================================================

-- Returns a shallow copy of the specified table.
-- @param t - the table to copy
-- @param copy - the table to copy into [optional]
local function tcopy(t, copy)
  if not copy then copy = _DBL:Get() end
  for k, v in pairs(t) do copy[k] = v end
  return copy
end

-- Bag iterator function.
local function iterateBags()
  local bag, slot = BACKPACK_CONTAINER, 0
  local numSlots = GetContainerNumSlots(bag)
  return function()
    slot = slot + 1
    if (slot > numSlots) then
      slot = 1
      repeat
        bag = bag + 1
        if (bag > NUM_BAG_SLOTS) then return nil end
        numSlots = GetContainerNumSlots(bag)
      until (numSlots > 0)
    end
    return bag, slot, GetContainerItemID(bag, slot)
  end
end

-- ============================================================================
-- _DBL Pooling Functions
-- ============================================================================

do
  local pool = {}

  -- Returns a table from the pool.
  function _DBL:Get()
    local t = next(pool)
    if t then pool[t] = nil else t = {} end
    return t
  end
  
  -- Releases a table into the pool.
  -- @param t - the table to release
  function _DBL:Release(t)
    self:Clean(t)
    pool[t] = true
  end

  -- Recursively cleans a table, releasing any child tables into the pool.
  -- @param t - the table to clean
  function _DBL:Clean(t)
    for k, v in pairs(t) do
      if (type(v) == "table") then self:Release(v) end
      t[k] = nil
    end
  end
end

-- ============================================================================
-- _DBL Tooltip Scanning Functions (copied from DethsTooltipLib)
-- ============================================================================

do
  -- Consts
  local ITEM_LEVEL_CAPTURE = ITEM_LEVEL:gsub("%%d", "(%%d+)")

  local tooltip = CreateFrame("GameTooltip", PREFIX.."Scanner", nil, "GameTooltipTemplate")
  local tooltipTextLeft = PREFIX.."ScannerTextLeft"
  local tooltipTextRight = PREFIX.."ScannerTextRight"

  -- Returns all non-blank lines in the scanner tooltip.
  -- @param rightSide - true for right-side scanning [optional]
  local function getAllLines(rightSide)
    local textSide = rightSide and tooltipTextRight or tooltipTextLeft
    local lines = _DBL:Get()

    for i = 1, tooltip:NumLines() do
      local line = strtrim((_G[textSide..i]):GetText()) or ""
      if (line ~= "") then lines[#lines+1] = line end
    end

    return lines
  end

  -- Returns true if tooltip information is available to be scanned.
  local function isScannable()
    local allLines = getAllLines()
    if (#allLines == 0) then return false end

    for _, line in pairs(allLines) do
      if (line == RETRIEVING_ITEM_INFO) then
        _DBL:Release(allLines)
        return false
      end
    end

    _DBL:Release(allLines)
    return true
  end

  -- Returns the first match for a specified pattern in the tooltip.
  -- @param rightSide - true for right-side scanning
  -- @param pattern - the pattern to match
  local function match(rightSide, pattern)
    local allLines = getAllLines(rightSide)
    local result
    
    for _, line in pairs(allLines) do
      result = line:match(pattern)
      if result then
        _DBL:Release(allLines)
        return result
      end
    end

    _DBL:Release(allLines)
    return nil
  end

  -- Scans and returns the bag item's item level from its tooltip.
  -- @param item - the bag item to scan
  function _DBL:ScanItemLevel(item)
    tooltip:SetOwner(UIParent, "ANCHOR_NONE")
    tooltip:SetBagItem(item.Bag, item.Slot)
    if isScannable() then
      return tonumber(match(false, ITEM_LEVEL_CAPTURE) or "") or item.ItemLevel
    end
  end
end

-- ============================================================================
-- _DBL.BagUpdateFrame
-- ============================================================================

do
  local bagUpdateFrame = CreateFrame("Frame", PREFIX.."BagUpdateFrame")
  local events = {
    BAG_UPDATE = true,
    PLAYER_LEVEL_UP = true,
    ACTIVE_TALENT_GROUP_CHANGED = true
  }

  function bagUpdateFrame:OnEvent(event, ...)
    if events[event] then _DBL:QueueUpdate() end
  end
  bagUpdateFrame:SetScript("OnEvent", bagUpdateFrame.OnEvent)
  for k in pairs(events) do bagUpdateFrame:RegisterEvent(k) end

  function bagUpdateFrame:OnUpdate(elapsed) _DBL:OnUpdate(elapsed) end
  bagUpdateFrame:SetScript("OnUpdate", bagUpdateFrame.OnUpdate)
end

-- ============================================================================
-- _DBL Functions (Private)
-- ============================================================================

-- Queues DBL for updating.
function _DBL:QueueUpdate()
  self.UpdateTimer = 0
  self.IsUpToDate = false
  self.UpdateQueued = true
  self.ListenersQueued = false
  for k in pairs(self.ItemLevelQueue) do self.ItemLevelQueue[k] = nil end
end

-- Called by bagUpdateFrame above.
function _DBL:OnUpdate(elapsed)
  -- Update items if queued and all item info is available
  if self.UpdateQueued then
    if self:VerifyCache() then
      self.UpdateTimer = 0
      self.UpdateQueued = false
      self.ListenersQueued = true
      self:UpdateItems()
    end
  -- Update queued item levels
  elseif next(self.ItemLevelQueue) then
    for item in pairs(self.ItemLevelQueue) do
      local itemLevel = _DBL:ScanItemLevel(item)
      if itemLevel then
        -- debug(format("%s ilvl: %s", item.ItemLink, itemLevel))
        item.ItemLevel = itemLevel
        self.ItemLevelQueue[item] = nil
      end
    end
  -- Notify listeners if queued and 1 second passes without a bag update
  elseif self.ListenersQueued then
    self.UpdateTimer = self.UpdateTimer + elapsed
    if (self.UpdateTimer >= 1) then -- 1 second delay
      if self:ValidateItems() then
        self.UpdateTimer = 0
        self.ListenersQueued = false
        for listener in pairs(self.Listeners) do listener() end
      end
    end
  end
end

-- Returns true if all items in the player's bags have info available.
function _DBL:VerifyCache()
  for bag, slot, itemID in iterateBags() do
    if itemID then -- slot is not empty
      --[[
        NOTE: Mythic keystones and battle pets use a different type of item
        link than all other items. These links cause GetItemInfo(link) to
        return empty. When this happens, we fall back to using
        GetItemInfo(itemID).
      ]]
      local link = GetContainerItemLink(bag, slot)
      local name = link and GetItemInfo(link) or GetItemInfo(itemID)
      if not name or (name == RETRIEVING_ITEM_INFO) then return false end
    end
  end
  return true
end

-- Returns true if all _DBL.Bags items are valid.
function _DBL:ValidateItems()
  for bag, slot, itemID in iterateBags() do
    local item = self.Bags[bag][slot]
    --[[
      Fail if:
      1. Bag slot has item id, be we have no item
      2. Our item does not have the same item id
    --]]
    if (not item and itemID) or (item and (itemID ~= item.ItemID)) then
      self:QueueUpdate()
      return false
    end
  end
  -- Items are valid
  return true
end

-- Updates the _DBL.Bags and _DBL.Items tables.
function _DBL:UpdateItems()
  for bag, slot, itemID in iterateBags() do
    -- Release item if slot is now empty
    if not itemID then
      if self.Bags[bag][slot] then
        local item = self.Bags[bag][slot]
        -- debug(format("[%s, %s] is now empty. Releasing %s.", bag, slot, item.ItemLink))
        self:Release(self.Bags[bag][slot])
        self.Bags[bag][slot] = nil
      end
    -- Otherwise, add item to bag table
    else
      -- local wasEmpty = false
      -- if not self.Bags[bag][slot] then wasEmpty = true end
      local item = self:GetItem(bag, slot, self.Bags[bag][slot])
      if item then
        self.Bags[bag][slot] = item
        -- if wasEmpty then debug(format("Set [%s, %s] to %s", bag, slot, item.ItemLink)) end
      else -- this should never occur, but just in case...
        self:QueueUpdate()
        return
      end
    end

    self.IsUpToDate = true
  end
end

-- ============================================================================
-- _DBL:GetItem()
-- ============================================================================

do
  -- Consts
  local ARMOR_CLASS = GetItemClassInfo(LE_ITEM_CLASS_ARMOR)
  local WEAPON_CLASS = GetItemClassInfo(LE_ITEM_CLASS_WEAPON)

  -- ========================================================
  -- Item Mixins
  -- ========================================================

  local Mixins = {}

  -- Returns true if the item is locked.
  function Mixins:IsLocked()
    local locked = select(3, GetContainerItemInfo(self.Bag, self.Slot))
    return locked
  end

  -- Returns true if the item has been filtered out by the blizzard bag search bar.
  function Mixins:IsFiltered()
    local filtered = select(8, GetContainerItemInfo(self.Bag, self.Slot))
    return filtered
  end

  -- ========================================================
  -- Main Function
  -- ========================================================

  -- Returns a table of item info for the item residing within a specified bag
  -- and slot, or nil if the info could not be retrieved.
  function _DBL:GetItem(bag, slot, t)
    -- Get item info, return nil if missing info
    -- Ignoring: locked (3), filtered (8)
    local texture, quantity, _, quality, readable, lootable, itemLink, _, noValue, itemID = GetContainerItemInfo(bag, slot)
    if not (texture and quantity and quality and (readable ~= nil) and (lootable ~= nil) and itemLink and (noValue ~= nil) and itemID) then return nil end

    -- Get additional item info, return nil if missing info
    -- Ignoring: itemLink (2), quality (3), itemLevel (4), texture (10)
    local name, _, _, _, reqLevel, class, subClass, maxStack, equipSlot, _, price, classID, subClassID, bindType, expacID, setID, isCraftingReagent = GetItemInfo(itemLink)
    if not name then name, _, _, _, reqLevel, class, subClass, maxStack, equipSlot, _, price, classID, subClassID, bindType, expacID, setID, isCraftingReagent = GetItemInfo(itemID) end
    if not (name and reqLevel and class and subClass and maxStack and equipSlot and price and classID and subClassID and bindType and expacID) then return nil end

    local itemLevel = GetDetailedItemLevelInfo(itemLink) or GetDetailedItemLevelInfo(itemID) or 1
    
    -- Build and return item
    local item = t or self:Get()
    item.Bag = bag
    item.Slot = slot

    -- Mixins
    for k, v in pairs(Mixins) do item[k] = v end

    -- GetContainerItemInfo
    item.Texture = texture
    item.Quantity = quantity
    item.Quality = quality
    item.Readable = readable
    item.Lootable = lootable
    item.ItemLink = itemLink
    item.NoValue = noValue
    item.ItemID = itemID

    -- GetItemInfo
    item.Name = name
    item.RequiredLevel = reqLevel
    item.Class = class
    item.SubClass = subClass
    item.MaxStack = maxStack
    item.EquipSlot = equipSlot
    item.Price = price
    item.ClassID = classID
    item.SubClassID = subClassID
    item.BindType = bindType
    item.ExpacID = expacID
    item.SetID = setID
    item.IsCraftingReagent = isCraftingReagent
    
    -- GetDetailedItemLevelInfo
    item.ItemLevel = itemLevel

    -- If item is armor or weapon, item level may be inaccurate. Queue for update via tooltip.
    if (class == ARMOR_CLASS) or (class == WEAPON_CLASS) then self.ItemLevelQueue[item] = true end

    return item
  end
end

-- ============================================================================
-- General Functions
-- ============================================================================

-- Registers a function to be called each time DBL fully updates.
-- @param listener - the listener function to add
function DBL:AddListener(listener)
  assert(type(listener) == "function", "listener must be a function")
  _DBL.Listeners[listener] = true
end

-- Removes a listener by reference.
-- @param listener - the listener function to remove
function DBL:RemoveListener(listener)
  if _DBL.Listeners[listener] then _DBL.Listeners[listener] = nil end
end

-- Returns true if all items are available and up-to-date.
function DBL:IsUpToDate()
  return _DBL.IsUpToDate
end

-- Returns true if the specified bag and slot does not contain an item.
-- @param bag - the bag index
-- @param slot - the slot index
function DBL:IsEmpty(bag, slot)
  return (GetContainerItemID(bag, slot) == nil)
end

-- Returns true if the specified item still resides in its associated bag slot.
-- @param item - the item
function DBL:StillInBags(item)
  local _, quantity, _, _, _, _, itemLink, _, _, _ = GetContainerItemInfo(item.Bag, item.Slot)
  return (item.Quantity == quantity) and (item.ItemLink == itemLink)
end

-- Returns true if two items are identical.
-- @param item1 - the first item
-- @param item2 - the second item
function DBL:Compare(item1, item2)
  for k, v in pairs(item1) do
    if not (item2[k] == v) then
      return false
    end
  end

  return true
end

-- Wipes and releases the specified table for reuse by DBL. This function
-- should only be used when the specified table will no longer be referenced.
-- @param t - the table to release
function DBL:Release(t)
  assert(type(t) == "table", "t must be a table")
  _DBL:Release(t)
end

-- ============================================================================
-- Item Functions
-- ============================================================================

-- Filters the specified item table using a specified filter function.
-- A filter function takes an item as a parameter and returns true or false.
-- Only items which cause the function to return true will remain in the table.
-- @param items - the table of items to filter
-- @param filterFunc - the filter function
function DBL:FilterItems(items, filterFunc)
  assert(type(filterFunc) == "function", "filterFunc must be a function")

  for i = #items, 1, -1 do
    if not filterFunc(items[i]) then
      _DBL:Release(tremove(items, i))
    end
  end
end

--[[
  This function returns one of the following:
    1. nil if the specified bag slot is empty
    2. The specified table after cleaning and coping item data into it
    3. A new table with up-to-date item data

  @param bag - the bag index
  @param slot - the slot index
  @param t - the table to copy item data into [optional]
]]
function DBL:GetItem(bag, slot, t)
  -- Verify item exists
  local item = _DBL.Bags[bag] and _DBL.Bags[bag][slot]
  if not item then return nil end
  if (type(t) == "table") then _DBL:Clean(t) end
  return tcopy(item, t)
end

-- This function returns either a new table with all up-to-date items, or the
-- specified table after cleaning and coping items into it.
-- @param t - the table to copy items into [optional]
function DBL:GetItems(t)
  -- Initialize t
  if (type(t) == "table") then _DBL:Clean(t) else t = _DBL:Get() end

  -- Copy
  for _, bag in pairs(_DBL.Bags) do
    for _, item in pairs(bag) do t[#t+1] = tcopy(item) end
  end

  return t
end

--[[
  This function is similar to DBL:GetItems(); however, the returned table will
  only contain items which cause a specified filter function to return true.
  
  Example filter function:
    local function filter(item)
      return item.Quality == LE_ITEM_QUALITY_POOR
    end
  
  @param filterFunc - the filter function
  @param t - the table to copy items into [optional]
  @param maxItems - the maximum number of items to copy [optional]
]]
function DBL:GetItemsByFilter(filterFunc, t, maxItems)
  assert(type(filterFunc) == "function", "filterFunc must be a function")
  if maxItems then
    assert((type(maxItems) == "number") and (maxItems > 0), "maxItems must be a number > 0")
  end

  -- Initialize t
  if (type(t) == "table") then _DBL:Clean(t) else t = _DBL:Get() end

  -- Copy
  for _, bag in pairs(_DBL.Bags) do
    for _, item in pairs(bag) do
      local copy = tcopy(item)
      if filterFunc(copy) then t[#t+1] = copy else _DBL:Release(copy) end
      if maxItems and (#t >= maxItems) then return t end
    end
  end

  return t
end
