-- Creator: contains simple create functions for UIObjects (frames, buttons, textures, etc.).

local LibName, LibVersion = "DethsFrameLib", "1.1"

-- DFL
local DFL = DethsLibLoader(LibName, LibVersion)
if not DFL.__initial_load then return end
local Creator = DFL.Creator

-- Wow upvalues
local CreateFrame, UIParent = CreateFrame, UIParent

-- ============================================================================
-- Frame
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_Frame_", LibName, LibVersion)

  -- Returns a frame.
  -- @param parent - the parent frame
  -- @return - a basic frame
  function Creator:CreateFrame(parent)
    count = count + 1

    local frame = CreateFrame("Frame", name..count, parent or UIParent)
    frame:Show()
    
    return frame
  end
end

-- ============================================================================
-- Button
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_Button_", LibName, LibVersion)

  -- Returns a button.
  -- @param parent - the parent of the button
  -- @return - a basic button
  function Creator:CreateButton(parent)
    count = count + 1

    local button = CreateFrame("Button", name..count, parent)
    button:RegisterForClicks("LeftButtonUp")
    button:Show()
    
    return button
  end
end

-- ============================================================================
-- Check Button
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_CheckButton_", LibName, LibVersion)
  local checkedTexture = "Interface\\Buttons\\UI-CheckBox-Check"
  local disabledTexture = "Interface\\Buttons\\UI-CheckBox-Check-Disabled"  

  -- Returns a check button.
  -- @param parent - the parent of the check button
  -- @return - a basic check button
  function Creator:CreateCheckButton(parent)
    count = count + 1

    local checkButton = CreateFrame("CheckButton", name..count, parent)
    checkButton:SetCheckedTexture(checkedTexture)
    checkButton:SetDisabledCheckedTexture(disabledTexture)
    checkButton:Show()

    return checkButton
  end
end

-- ============================================================================
-- Texture
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_Texture_", LibName, LibVersion)

  -- Returns a texture.
  -- @param parent - the parent frame
  -- @param layer - the draw layer ("ARTWORK", "BACKGROUND", etc.)
  -- @return - a basic texture
  function Creator:CreateTexture(parent, layer)
    count = count + 1
    
    local texture = parent:CreateTexture(name..count, (layer or "BACKGROUND"))
    texture:SetAllPoints()
    texture:Show()

    return texture
  end
end

-- ============================================================================
-- Font String
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_FontString_", LibName, LibVersion)

  -- Returns a font string.
  -- @param parent - the parent frame
  -- @param text - the initial text [optional]
  -- @param font - the font style to inherit [optional]
  -- @param layer - the draw layer ("ARTWORK", "BACKGROUND", etc.) [optional]
  -- @return - a basic font string
  function Creator:CreateFontString(parent, text, font, layer)
    count = count + 1
    
    local fontString = parent:CreateFontString(name..count, (layer or "OVERLAY"), (font or DFL.Fonts.Normal))
    fontString:SetText(text)
    fontString:Show()

    return fontString
  end
end

-- ============================================================================
-- Scroll Frame
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_ScrollFrame_", LibName, LibVersion)

  -- Returns a scroll frame.
  -- @param parent - the parent frame
  -- @return - a basic scroll frame
  function Creator:CreateScrollFrame(parent)
    count = count + 1
    
    local scrollFrame = CreateFrame("ScrollFrame", name..count, parent)
    scrollFrame:SetClipsChildren(true)
    scrollFrame:Show()

    return scrollFrame
  end
end

-- ============================================================================
-- Slider
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_Slider_", LibName, LibVersion)

  -- Returns a slider.
  -- @param parent - the parent frame
  -- @return - a basic slider
  function Creator:CreateSlider(parent)
    count = count + 1
    
    local slider = CreateFrame("Slider", name..count, parent)
    slider:SetObeyStepOnDrag(true)
    slider:SetMinMaxValues(0, 0)
    slider:SetValue(0)
    slider:Show()

    return slider
  end
end

-- ============================================================================
-- Edit Box
-- ============================================================================

do
  local count = 0
  local name = format("%s_%s_EditBox_", LibName, LibVersion)
  local EditBox_ClearFocus = EditBox_ClearFocus

  -- Returns an edit box.
  -- @param parent - the parent frame
  -- @param maxLetters - the maximum amount of characters [optional]
  -- @param numeric - whether or not the edit box only accepts numeric input [optional]
  -- @param font - the font style for the edit box to inherit [optional]
  -- @return - a basic edit box
  function Creator:CreateEditBox(parent, maxLetters, numeric, font)
    count = count + 1

    local editBox = CreateFrame("EditBox", name..count, parent)
    editBox:SetScript("OnEscapePressed", EditBox_ClearFocus)
    editBox:SetFontObject(font or DFL.Fonts.Normal)
    editBox:SetMaxLetters(maxLetters or 0)
    editBox:SetNumeric(numeric)
    editBox:SetMultiLine(false)
    editBox:SetAutoFocus(false)
    editBox:ClearFocus()

    editBox:Show()

    return editBox
  end
end
