-- DFL
local DFL = DethsLibLoader("DethsFrameLib", "1.1")
if not DFL.__initial_load then return end

local Points = DFL.Points

-- Upvalues
local assert, floor, pairs, ipairs, max, nop, tonumber, type, unpack =
      assert, floor, pairs, ipairs, max, nop, tonumber, type, unpack

-- FauxScrollFrame
local FauxScrollFrame = DFL.FauxScrollFrame
FauxScrollFrame.Scripts = {}
FauxScrollFrame.SliderScripts = {}

-- ============================================================================
-- Creation Function
-- ============================================================================

-- Creates and returns a DFL faux scroll frame.
-- @param parent - the parent frame
-- @param data - a table of data for faux objects. Data items are retrieved by
-- index and passed to objects using an object's SetData(data) function.
-- @param objFunc - the function called to create faux scroll objects. These
-- objects must implement a SetData(data) function.
-- @param numObjects - the number of objects to create
-- @param sliderSpacing - spacing between the frame and slider [optional]
-- @param objSpacing - padding and spacing for the objects [optional]
function FauxScrollFrame:Create(parent, data, objFunc, numObjects, sliderSpacing, objSpacing)
  assert(type(data) == "table", "data must be a table")
  assert(type(objFunc) == "function", "objFunc must be a function")
  numObjects = tonumber(numObjects) or 0
  assert(numObjects > 0, "numObjects must be > 0")
  
  local frame = DFL.Creator:CreateFrame(parent)
  frame._sliderSpacing = tonumber(sliderSpacing) or DFL:Padding(0.25)
  frame._objSpacing = tonumber(objSpacing) or DFL:Padding(0.5)
  frame._data = data
  frame._offset = 0

  -- objFrame, holds objects created with objFunc
  local objFrame = DFL.Creator:CreateFrame(frame)
  objFrame._texture = DFL.Texture:Create(objFrame)
  objFrame._objects = {}
  frame._objFrame = objFrame
  
  do -- Add required number of objects
    local lastObj
    local minWidth = 0
    local minHeight = 0

    for i = 1, numObjects do
      local obj = objFunc(objFrame)
      minWidth = max(minWidth, obj:GetWidth())
      minHeight = minHeight + obj:GetHeight()

      if not (type(obj.SetData) == "function") then
        error("FauxScrollFrame objects must have a SetData function")
      end

      if lastObj then
        obj:SetPoint(Points.TOPLEFT, lastObj, Points.BOTTOMLEFT, 0, -frame._objSpacing)
        obj:SetPoint(Points.TOPRIGHT, lastObj, Points.BOTTOMRIGHT, 0, -frame._objSpacing)
      else
        obj:SetPoint(Points.TOPLEFT, frame._objSpacing, -frame._objSpacing)
        obj:SetPoint(Points.TOPRIGHT, -frame._objSpacing, -frame._objSpacing)
      end

      lastObj = obj
      objFrame._objects[#objFrame._objects+1] = obj
    end
    minHeight = minHeight + (#objFrame._objects+1) * frame._objSpacing
    objFrame._minWidth = minWidth + (frame._objSpacing * 2)
    objFrame:SetHeight(minHeight)
  end

  -- slider
  local slider = DFL.Slider:Create(frame)
  slider.SetEnabled = nop
  slider:SetValueStep(1)
  frame._slider = slider
  
  -- Mixins
  DFL:AddDefaults(frame)
  DFL:AddMixins(frame, self.Functions)
  DFL:AddScripts(frame, self.Scripts)
  DFL:AddScripts(slider, self.SliderScripts)

  frame:SetColors()
  frame:Refresh()

  return frame
end

-- ============================================================================
-- Functions
-- ============================================================================

do
  local Functions = FauxScrollFrame.Functions

  function Functions:SetData(data)
    assert(type(data) == "table", "data must be a table")
    self._data = data
    self._offset = 0
    self._slider:SetValue(0)
  end

  -- Enables or disables the faux scroll frame.
  function Functions:OnSetEnabled(enabled)
    for _, obj in pairs(self._objFrame._objects) do
      if obj.SetEnabled then obj:SetEnabled(enabled) end
    end
  end

  function Functions:SetColors(color, sliderColors, ...)
    self._objFrame._texture:SetColors(color or DFL.Colors.Area)
    if sliderColors then self._slider:SetColors(unpack(sliderColors)) end
    if (select("#", ...) > 0) then
      for _, object in pairs(self._objFrame:GetChildren()) do
        if object.SetColors then object:SetColors(...) end
      end
    end
  end

  -- Refreshes the faux scroll frame.
  function Functions:Refresh()
    for _, obj in pairs(self._objFrame._objects) do
      if obj.Refresh then obj:Refresh() end
    end
    self._objFrame._texture:Refresh()
    self._slider:Refresh()
  end

  function Functions:Resize()
    local objFrame = self._objFrame
    local slider = self._slider
    -- Resize slider
    slider:Resize()
    -- Resize frame
    local minWidth = objFrame._minWidth + (self._sliderSpacing + slider:GetWidth())
    self:SetWidth(minWidth)
    self:SetHeight(objFrame:GetHeight())
  end
end

-- ============================================================================
-- Scripts
-- ============================================================================

do
  local Scripts = FauxScrollFrame.Scripts

  function Scripts:OnUpdate(elapsed)
    local objFrame = self._objFrame
    local slider = self._slider
    local spacing = self._sliderSpacing
    local objects = self._objFrame._objects

    -- Update objects
    for i, object in ipairs(objects) do
      local index = (i + self._offset)
      local data = self._data[index]

      if data then
        object:Show()
        object:SetData(data)
      else
        object:Hide()
      end
    end

    -- Update slider values, show slider if max value is > 0
    local maxVal = max((#self._data - #objects), 0)
    objFrame:SetPoint(Points.TOPLEFT)
    if (maxVal > 0) then
      slider:SetMinMaxValues(0, maxVal)
      slider:Show()
      objFrame:SetPoint(Points.BOTTOMRIGHT, -(spacing + slider:GetWidth()), 0)
      slider:SetPoint(Points.TOPLEFT, objFrame, Points.TOPRIGHT, spacing, 0)
      slider:SetPoint(Points.BOTTOMLEFT, objFrame, Points.BOTTOMRIGHT, spacing, 0)
    else
      slider:SetMinMaxValues(0, 0)
      slider:Hide()
      objFrame:SetPoint(Points.BOTTOMRIGHT)
    end

    if self.OnUpdate then self:OnUpdate(elapsed) end
  end

  function Scripts:OnMouseWheel(delta)
    self._slider:SetValue(self._slider:GetValue() - delta)
  end
end

-- ============================================================================
-- SliderScripts
-- ============================================================================

do
  local SliderScripts = FauxScrollFrame.SliderScripts

  function SliderScripts:OnValueChanged()
    self:GetParent()._offset = floor(self:GetValue() + 0.5)
  end
end
