-- DFL
local DFL = DethsLibLoader("DethsFrameLib", "1.1")
if not DFL.__initial_load then return end

-- Upvalues
local UIParent, UISpecialFrames = UIParent, UISpecialFrames
local pairs, unpack = pairs, unpack

-- ParentFrame
local ParentFrame = DFL.ParentFrame
ParentFrame.FrameFunctions = {}
ParentFrame.FrameScripts = {}

-- ============================================================================
-- Creation Function
-- ============================================================================

-- Creates and returns a DFL parent frame.
function ParentFrame:Create()
  local parentFrame = DFL.ChildFrame:Create()
  DFL:AddMixins(parentFrame, self.Functions)
  return parentFrame
end

-- ============================================================================
-- Functions
-- ============================================================================

do
  local Functions = ParentFrame.Functions

  -- Initializes the frame.
  function Functions:Initialize()
    -- Create underlying frame
    local frame = DFL.Frame:Create()
    self.Frame = frame
    frame._parent_frame = self
    -- Center
    frame:SetPoint("CENTER")
    frame:SetFrameStrata("FULLSCREEN_DIALOG")
    frame:SetToplevel(true)
    -- Make draggable
    frame:SetClampedToScreen(true)
    frame:EnableMouse(true)
    frame:SetMovable(true)
    frame:RegisterForDrag("LeftButton")
    -- Allow WoW to handle hiding frame when necessary
    UISpecialFrames[#UISpecialFrames+1] = frame:GetName()
    
    DFL:AddMixins(frame, ParentFrame.FrameFunctions)
    DFL:AddScripts(frame, ParentFrame.FrameScripts)
    
    frame:SetDirection(DFL.Directions.DOWN)
    frame:SetMinWidth(10)
    frame:SetMinHeight(10)
    frame:SetColors()
    frame:Refresh()
    
    self:OnInitialize()
    frame:Hide()
    
    -- Nop so the above code is only executed once
    self.Initialize = nop
    self.OnInitialize = nop
  end
  Functions.OnInitialize = nop

  -- Returns the title ChildFrame of the ParentFrame.
  function Functions:GetTitle()
    return self._title
  end

  -- Returns the content ChildFrame of the ParentFrame.
  function Functions:GetContent()
    return self._content
  end

  do -- SetTitle(), SetContent()
    local function addChild(self, child, index)
      if child then
        child:Initialize()
        child:Show()
        child:SetEnabled(self:IsEnabled())
        self.Frame:Add(child.Frame, index)
      end

      --[[
        NOTE: Certain objects (ScrollFrame/FauxScrollFrame) are not resized or
        displayed properly within the Title/Content child frames unless we
        call self.Frame:Resize() an extra time here. I don't know why.

        Also, note that that the ParentFrame's Frame will be queued for resize
        if either of the calls to Frame:Add() or Frame:Remove() are executed.
        
        It's a band-aid, but the perfomance impact is negligible.
      --]]
      self.Frame:Resize()
      self.Frame:Resize()
    end

    -- Sets the title ChildFrame of the ParentFrame.
    function Functions:SetTitle(title)
      if self._title then self.Frame:Remove(self._title.Frame) end
      if title then assert(DFL:IsType(title, DFL.ChildFrame:GetType()), "title must be a ChildFrame") end
      self._title = title
      addChild(self, title, 1)
    end

    -- Sets the content ChildFrame of the ParentFrame.
    function Functions:SetContent(content)
      if self._content then self.Frame:Remove(self._content.Frame) end
      if content then assert(DFL:IsType(content, DFL.ChildFrame:GetType()), "content must be a ChildFrame") end
      self._content = content
      addChild(self, content)
    end
  end

  -- Queues the ParentFrame's frame for resizing.
  function Functions:QueueResize()
    self.Frame._resize_queued = true
  end
end

-- ============================================================================
-- FrameFunctions
-- ============================================================================

do
  local Functions = ParentFrame.FrameFunctions

  -- Returns the ParentFrame the frame belongs to.
  function Functions:GetParent()
    return self._parent_frame
  end

  function Functions:SetColors(color, borderColor)
    DFL.Frame.Functions.SetColors(self, color)
    self._borderColor = borderColor or self._borderColor or DFL.Colors.Area
  end

  function Functions:Refresh()
    DFL.Frame.Functions.Refresh(self)
    DFL:AddBorder(self, unpack(self._borderColor))
  end
end

-- ============================================================================
-- FrameScripts
-- ============================================================================

do
  local Scripts = ParentFrame.FrameScripts
  
  function Scripts:OnDragStart() self:StartMoving() end
  function Scripts:OnDragStop() self:StopMovingOrSizing() end

  function Scripts:OnUpdate(elapsed)
    if self._resize_queued then
      self._resize_queued = false
      self:Resize()
    end
    
    if self._parent_frame.OnUpdate then
      self._parent_frame:OnUpdate(elapsed)
    end
  end
end
